# python standard imports
from argparse import ArgumentParser
from json import dumps
from pathlib import Path
from re import sub

# internal imports
from .. import Citator
from .resources import unify_regex, format_page, sources_table
from .resources import SOURCES_INTRO, VERSION

COPYRIGHT_MESSAGE = f"""
// This script was made with CiteURL {VERSION}, an extensible framework
// to turn legal references into URLs.
//
// The "templates" variable below holds the data necessary to 
// turn each kind of citation into a URL. Some or all of the templates may
// have been made by a third party and are not part of CiteURL itself.
//
// CiteURL is copyright of Simon Raindrum Sherred under the MIT License.
// See https://raindrum.github.io/citeurl for more info.
"""

_dir = Path(__file__).parent.absolute()
BASE_JS_PATH =  _dir / 'citeurl.js'

PAGE = """
<h1>CiteURL</h1><title>CiteURL</title>
<div class="narrow">
<p>Type a legal citation into the box below, and I'll try to send you
directly to the case or law that it references:</p>
<form onsubmit="handleSearch(event)">
<div class="searchbar">
  <input type="search" name="q" id="q" placeholder="Enter citation..."
  label="Citation search bar"><button type="submit">Go</button>
</div>
<p>
  <label for="q" id="explainer" class="explainer"> </label>
</p>
</form>
</div>
{sources_table}
"""

# these functions either uncomment or remove a commented-out code block
# that looks like this:
# /*blockname
# <whatever code goes here>
# */
def _uncomment(blockname: str, source_js: str):
    return sub(fr' */\*{blockname}\n([^©]+?) *\*/\n', r'\1', source_js)
def _remove(blockname: str, source_js: str):
    return sub(fr' */\*{blockname}\n[^©]+? *\*/\n', '', source_js)

def makejs(
    citator: Citator,
    entire_page: bool = False,
    include_sources_table: bool = False,
) -> str:
    """
    Generate a JavaScript implementation of a given citator's lookup
    features, so that it can be embedded in a website. Optionally
    package it as a standalone web page for end users.
    
    Arguments:
        citator: a CiteURL citator object, with any number of templates
            loaded.
        entire_page: whether to output an HTML page with a searchbar
            and styling
        include_sources_table: whether to provide a table listing each
            template that the page supports. Implies entire_page = True.
    Returns:
        a string containing JavaScript or HTML
    """
    if include_sources_table:
        entire_page = True
    
    # translate each template to json
    json_templates = []
    for template in citator.templates.values():
        if not template.URL_builder:
            continue
        
        json = {}
        
        # some parts of a template can be copied over easily
        json['name'] = template.name
        
        defaults = template.meta
        for name, token in template.tokens.items():
            if token.default:
                defaults[name] = token.default
        if defaults:
            json['defaults'] = defaults
        
        json['regexes'] = [
            r.pattern.replace('?P<', '?<') for r in template.broad_regexes
        ]

        # only add the relevant information from each operation
        
        json['operations'] = []
        
        for name, token in template.tokens.items():
            for edit in token.edits:
                edit_dict = {
                    'token': name,
                    edit.action: edit.data
                }
                if 'lookup' in edit_dict and not edit.mandatory:
                    edit_dict['optionalLookup'] = edit_dict.pop('lookup')
                json['operations'].append(edit_dict)
        
        for edit in template.URL_builder.edits:
            edit_dict = {
                'token': edit.token,
                edit.action: edit.data
            }
            if edit.output:
                edit_dict['output'] = edit.output
            if 'lookup' in edit_dict and not edit.mandatory:
                edit_dict['optionalLookup'] = edit_dict.pop('lookup')
            json['operations'].append(edit_dict)
        
        json['URL'] = [p for p in template.URL_builder.parts]

        json_templates.append(json)
    
    # write json to str
    json_str = dumps(
        json_templates,
        indent=4,
        sort_keys=False,
        ensure_ascii=False,
    )
    
    # generate javascript
    javascript = (
        COPYRIGHT_MESSAGE
        + '\nconst templates = ' 
        + json_str + ';\n\n'
        + BASE_JS_PATH.read_text()
    )
    
    # uncomment or remove browser-only features in the JS
    if entire_page:
        javascript = _uncomment('PAGEBEHAVIOR', javascript)
        javascript = _uncomment('LOGS', javascript)
    else:
        javascript = _remove('LOGS', javascript)
        javascript = _remove('PAGEBEHAVIOR', javascript)
    
    if include_sources_table:
        table = (f'{SOURCES_INTRO}\n{sources_table(citator)}')
    else:
        table = ''
    
    # optionally embed the javascript into an HTML page
    if entire_page:
        output = format_page(
            PAGE,
            sources_table=table,
            inline_logo=True,
            inline_css=True,
            js=javascript,
            relation='Static page generated with',
        )
    else:
        output = javascript
    
    return output



import re
from typing import Union
from functools import cached_property
from copy import copy

from .citator import list_cites
from .citation import Citation

class Authority:
    def __init__(
        self,
        model_cite: Citation,
        ignored_tokens = ['subsection', 'clause', 'pincite', 'paragraph'],
    ):
        self.template = model_cite.template
        self.ignored_tokens = ignored_tokens
        self.tokens = {}
        for key, value in model_cite.tokens.items():
            if key in ignored_tokens:
                break
            else:
                self.tokens[key] = value
        self.citations = [model_cite]
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return (
            f'{self.name} ({len(self.citations)} '
            f'cite{"s" if len(self.citations) > 1 else ""})'
        )
    
    def __contains__(self, cite: Citation):
        if cite.template.name != self.template.name:
            return False
        for key, value in self.tokens.items():
            if value and cite.tokens.get(key) != value:
                if (
                    self.template.tokens[key].severable
                    and cite.tokens[key]
                    and cite.tokens[key].startswith(value)
                ):
                    continue
                else:
                    return False
        else:
            return True
    
    @cached_property
    def name(self):
        # build a name the proper way if a name_builder is defined
        if self.template.name_builder:
            return self.template.name_builder(self.tokens)
        
        # otherwise use horrible regex magic to reverse-engineer a name.
        # first find a longform citation to use as a starting point
        base_cite = self.citations[0]
        while base_cite.parent:
            base_cite = base_cite.parent
        
        # next construct a regex pattern to pull out the relevant
        # tokens, along with the non-token text preceding each one.
        # the non-token "prelude" text lets us replace the text of each
        # token only when it appears in the proper context
        pattern = ''
        for token in self.tokens:
            regex = re.escape(base_cite.tokens[token])
            segment = f"""((?P<{token}_prelude>.*?)(?P<{token}>{regex}))?"""
            if pattern:
                pattern = pattern[:-2] + segment + ')?'
            else:
                pattern = segment
        
        match = re.match(pattern, base_cite.text)
        
        # slice off all the text after the last relevant token. This is
        # to remove thingsl like subsections, etc. It assumes that all
        # the optional tokens (subsection, pincite, etc) appear *after*
        # all the mandatory ones.
        base_cite_text = base_cite.text[:match.span(token)[1]]
        
        # for each token, replace the value from the longform citation
        # with the corresponding value for *this* authority
        for token in self.tokens:
            if not match.group(token):
                continue
            prelude = str(match.group(f'{token}_prelude'))
            old_value = prelude + match.group(token)
            new_value = prelude + self.tokens[token]
            base_cite_text = base_cite_text.replace(old_value, new_value)
        return base_cite_text
    
    @cached_property
    def URL(self):
        if self.template.URL_builder:
            url =  self.template.URL_builder(self.tokens)
            if url:
                url = url.replace(' ', '%20')
        else:
            url = None
        return url
        

def list_authorities(
    source: Union[list[Citation], str],
    ignored_tokens = ['subsection', 'clause', 'pincite', 'paragraph'],
) -> list[Authority]:
    """
    Get a list of all the authorities that appear in the given list of
    citations. An authority represents a distinct section of law or
    court case. Two citations to the same authority can have different
    tokens, as long as those tokens are in the list of ignored_tokens
    """
    authorities = []
    if type(source) is str:
        source = list_cites(source)
    for cite in source:
        for authority in authorities:
            if cite in authority:
                authority.citations.append(cite)
                break
        else:
            authorities.append(Authority(cite, ignored_tokens))
    authorities.sort(key=lambda x: -len(x.citations))
    return authorities



"""
A Python-Markdown extension to detect citations and insert them into
the processed text as hyperlinks.
"""

# python standard imports
import re
import xml.etree.ElementTree as etree
from pathlib import Path

# markdown imports
from markdown.extensions import Extension
from markdown.postprocessors import Postprocessor

# internal imports
from . import Citator, insert_links
from .citator import _get_default_citator

# store citator in a global variable so it isn't remade each document
CITATOR: Citator = None

class CitationPostprocessor(Postprocessor):
    def __init__(
        self,
        citator,
        attributes: dict,
        redundant_links: bool,
        URL_optional: bool,
        break_id_on_regex: str,
        ignore_markup: bool,
    ):
        super().__init__()
        self.citator = citator
        self.attributes = attributes
        self.redundant_links = redundant_links
        self.URL_optional = URL_optional
        self.ignore_markup = ignore_markup,
        
        if break_id_on_regex:
            self.id_breaks = re.compile(break_id_on_regex)
        else:
            self.id_breaks = None
    
    def run(self, text):
        return insert_links(
            text = text,
            attrs = self.attributes,
            redundant_links = self.redundant_links,
            URL_optional = self.URL_optional,
            id_breaks = self.id_breaks,
            ignore_markup = self.ignore_markup,
            citator = self.citator,
        )

class CiteURLExtension(Extension):
    """Detects legal citations and inserts relevant hyperlinks."""
    def __init__(self, **kwargs):
        self.config = {
            'custom_templates': [
                [],
                'List of paths to YAML files containing additional citation'
                + 'templates to load. - Default: []',
            ],
            'use_defaults': [
                True,
                "Load CiteURL's default citation templates? - Default: True"
            ],
            'redundant_links': [
                False,
                (
                    "Whether to insert links links whose URLs are identical "
                    "to the previous URL"
                )
            ],
            'URL_optional': [
                False,
                (
                    "Whether to add <a> elements for citations that have "
                    "no URL"
                )
            ],
            'break_id_on_regex': [
                None,
                "Anywhere this string (parsed as regex) appears in the text, "
                + "chains of citations like 'id.' will be interrupted. Note "
                + "that this is based on the output HTML, *not* the original "
                + f"Markdown text."
            ],
            'attributes': [
                {'class': 'citation'},
                ("A dictionary of attributes (besides href) that the inserted"
                + " links should have. - Default: '{'class': 'citation'}'")
            ],
            'ignore_markup': [
                True,
                (
                    "Whether to detect citations even when they are "
                    'interrupted by inline markup, like "<i>Id.</i> at 32. '
                    'Default: True'
                )
            ],
        }
        super(CiteURLExtension, self).__init__(**kwargs)
    
    def extendMarkdown(self, md):
        global CITATOR
        if not CITATOR:
            if self.config['use_defaults'][0]:
                CITATOR = _get_default_citator()
            else:
                CITATOR = Citator(defaults=None)
        for path in self.config['custom_templates'][0] or []:
            CITATOR.load_yaml(Path(path).read_text())
        
        md.postprocessors.register(
            CitationPostprocessor(
                CITATOR,
                self.config['attributes'][0],
                self.config['redundant_links'][0],
                self.config['URL_optional'][0],
                self.config['break_id_on_regex'][0],
                self.config['ignore_markup'][0],
            ),
            "CiteURL",
            1
        )

def makeExtension(**kwargs):
    return CiteURLExtension(**kwargs)



# python standard imports
import re
from copy import copy


class TokenOperation:
    """A function to perform a predefined string manipulation"""
    def __init__(
        self,
        action: str,
        data,
        mandatory: bool = True,
        token: str = None,
        output: str = None,
    ):
        """
        Arguments:
            action: The kind of string manipulation that this operation
                will perform, using the given data. There are a few
                different options:
                
                'sub': Regex substitution to perform on the text. Needs
                    a list of two values: [PATTERN, REPLACEMENT]
                
                'lookup': Check if the token matches any of the given
                    regexes (via case-insensitive matching), and if so,
                    replace it with the corresponding value. Needs a
                    dictionary of `regex`: `replacement` pairs.
                
                'case': Capitalize the token in the specified way.
                    Options are 'upper', 'lower', and 'title'.
                
                'lpad': Left pad the token with zeros until it is the
                    specified number of characters long. Requires an
                    int specifying the number of characters. You can
                    also specify the padding character by providing a
                    tuple: (MINIMUM_LENGTH, PADDING_CHARACTER).
                
                'number_style': Assume that the token is a number,
                    either in the form of digits, Roman numerals, or
                    number words like "thirty-seven". Convert it into
                    the specified number format, which can be any of
                    these:
                    
                    'cardinal', e.g. "twenty-seven"
                    
                    'cardinal spaced', e.g. "twenty seven"
                    
                    'cardinal unspaced', e.g. "twentyseven"
                    
                    'ordinal', e.g. "twenty-seventh"
                    
                    'ordinal spaced', e.g. "twenty seventh"
                    
                    'ordinal unspaced', e.g. "twentyseventh"
                    
                    'roman numeral', e.g. 'xxvii'
                    
                    'digit', e.g. '27'
                    
                    Note that number formatting only works for positive
                    whole numbers that do not exceed 100.
                
            data: any data that a given action needs specified, as
                described above
            
            mandatory: whether a failed lookup or format action should
                invalidate the entire citation
            
            token: Necessary for operations in StringBuilders. This
                value lets you provide the name of input token to use,
                allowing you to then use the modify_dict() method.
            
            output: If this value is set, modify_dict() will save the
                operation's output to the dictionary key with this name
                instead of modifying the input token in place.
        """
        if action == 'sub':
            self.func = lambda x: re.sub(data[0], data[1], x)
        elif action == 'lookup':
            table = {
                re.compile(k, flags=re.I):v
                for k, v in data.items()
            }
            self.func = lambda x: self._lookup(x, table, mandatory)
        elif action == 'case':
            self.func = lambda x: self._set_case(x, data)
        elif action == 'lpad':
            self.func = lambda x: self._left_pad(x, data)
        elif action == 'number_style':
            action_options = ['cardinal', 'ordinal', 'roman', 'digit']
            if data not in action_options:
                raise SyntaxError(
                    f'{data} is not a valid number style. Valid options: '
                    f'{action_options}'
                )
            self.func = lambda x: self._number_style(x, data, mandatory)
        else:
            raise SyntaxError(
                f'{action} is not a defined token operation.'
            )
        
        self.action = action
        self.data = data
        self.mandatory = mandatory
        self.token = token
        self.output = output
    
    @classmethod
    def from_dict(cls, data: dict):
        "load a TokenOperation from a dictionary of values"
        operations = []
        for key in ['sub', 'lookup', 'case', 'lpad', 'number style']:
            value = data.get(key)
            if value:
                action = key.replace(' ', '_')
                action_data = value
                break
        mandatory = data.get('mandatory', True)
        token = data.get('token')
        output = data.get('output')
        return cls(action, action_data, mandatory, token, output)
    
    def to_dict(self) -> dict:
        "save this TokenOperation to a dictionary of values"
        output = {}
        for key in ['token', 'output']:
            if self.__dict__.get(key):
                output[key] = self.__dict__[key]
        output[self.action] = self.data
        if not self.mandatory:
            output['mandatory'] = False
        
        spaced_output = {k.replace('_', ' '):v for k, v in output.items()}
        
        return spaced_output
    
    def modify_dict(self, tokens: dict):
        """
        apply this operation to a dictionary of tokens,
        editing them as appropriate
        """
        if not tokens.get(self.token):
            return
        if self.output:
            tokens[self.output] = self.func(tokens[self.token])
        else:
            tokens[self.token] = self.func(tokens[self.token])
    
    def __call__(self, input_value):
        return self.func(input_value)
    
    def __repr__(self):
        return (
            f'TokenOperation(action="{self.action}", data="{self.data}"'
            + (f', mandatory=False' if not self.mandatory else '')
            + (f', token="{self.token}"' if self.token else '')
            + (f', output="{self.output}"' if self.output else '')
            + ')'
        )
    
    # ================ Token Processing Operations =================== #
    
    def _lookup(
        self,
        input: str,
        table: dict[re.Pattern, str],
        mandatory: bool=False,
    ) -> str:
        for pattern, repl in table.items():
            if pattern.fullmatch(input):
                return repl
        if mandatory:
            regexes = [r.pattern for r in table.keys()]
            raise SyntaxError(f'{input} could not be found in {table}')
        else:
            return input
    
    def _set_case(self, input: str, case: str) -> str:
        if case == 'upper':
            return input.upper()
        elif case == 'lower':
            return input.lower()
        elif case == 'title':
            return input.title()
    
    def _left_pad(self, input: str, min_length: int, pad_char='0'):
        diff = min_length - len(input)
        if diff > 0:
            return (pad_char*diff + input)
        return input
    
    def _number_style(self, input: str, form: str, throw_error: bool=False):
        if input.isnumeric():
            value = int(input)
        elif input[:-2].isnumeric(): # e.g. "2nd"
            value = int(input[:-2])
        else:
            input = input.lower()
            for i, row in enumerate(number_words):
                if input in row:
                    value = i + 1
                    break
            else:
                if throw_error:
                    raise SyntaxError(
                        f'{input} cannot be recognized as a number'
                    )
        if form == 'digit':
            return str(value)
        forms = ['roman', 'cardinal', 'ordinal']
        try:
            output = number_words[value - 1][forms.index(form)]
        except IndexError:
            return NotImplementedError(
                f"CiteURL cannot process a number as high as {value}"
            )
        if form == 'roman':
            return output.upper()
        return output


class TokenType:
    """
    These objects represent categories of tokens that might be found in
    a citation.
    
    Attributes:
        regex: A regular expression that matches the actual text of the
            token as found in any document, like the "42" in "42 USC §
            1983" or the "Fourteenth" in "The Fourteenth Amendment".
            This regex will automatically be enclosed in a named capture
            group and inserted into any of the template's match patterns
            wherever the token's name appears in curly braces.
        edits: Steps to normalize the token as captured in the regex
            into a value that is consistent across multiple styles.
        default: Set the token to this value if it is not found in the
            citation.
        severable: If two citations only differ based on this token,
            and only because one of the tokens extends longer than the
            other, e.g. "(b)(2)" and "(b)(2)(A)", then `severable` means
            that the former citation is thought to encompass the latter.
    """
    def __init__(
        self,
        regex: str = r'\d+',
        edits: list[TokenOperation] = [],
        default: str = None,
        severable: bool = False,
    ):
        self.regex = regex
        self.edits = edits
        self.default = default
        self.severable = severable
    
    @classmethod
    def from_dict(cls, name: str, data: dict):
        "load a TokenType from a dictionary of values"
        return cls(
            regex = data['regex'],
            default = data.get('default'),
            edits = [
                TokenOperation.from_dict(v)
                for v in data.get('edits', [])
            ],
            severable=data.get('severable', False)
        )
    
    def to_dict(self) -> dict:
        "save this TokenType to a dictionary for storage in YAML format"
        output = {'regex': self.regex}
        if self.edits:
            output['edits'] = [
                e.to_dict() for e in self.edits
            ]
        if self.default:
            output['default'] = self.default
        if self.severable:
            output['severable'] = True
        return output
    
    def normalize(self, token: str) -> str:
        if not token:
            return self.default
        for op in self.edits:
            token = op(token)
        return token
    
    def __str__(self):
        return self.regex
    
    def __repr__(self):
        norms = '[' + ', '.join([
            repr(n) for n in self.edits or []
        ]) + ']'
        return (
            f"TokenType(regex='{self.regex}'"
            + (f", default='{self.default}'" if self.default else '')
            + (f', edits={norms}' if self.edits else '')
            + ')'
        )


class StringBuilder:
    """
    A function to take a dictionary of values and use it to construct a
    piece of text from them. This is used for citation templates' name
    builders and URL builders. 
    
    Attributes:
        parts: A list of strings that will be concatenated to create the
            string. Parts may contain bracketed references to citations'
            token values as well as templates' metadata. If a part
            references a token whose value is not set, the part will be
            omitted from the created string.
        edits: A list of TokenOperations that will be performed on the
            provided tokens before the string is constructed. If the
            edits have `output` values, it is possible for them to
            define entirely new tokens for the sole purpose of building
            the string.
        defaults: A dictionary of default token values to use when not
            overwritten by the citation. Generally these are provided by
            the template's meta attribute.
    """
    def __init__(
        self,
        parts: list[str],
        edits: list[TokenOperation] = [],
        defaults: dict[str, str] = {}
    ):
        self.parts = parts
        self.edits = edits
        self.defaults = defaults
    
    @classmethod
    def from_dict(cls, data: dict):
        "load StringBuilder from dictionary of values"
        edits = [
            TokenOperation.from_dict(o)
            for o in data.get('edits', [])
        ]
        parts = data['parts']
        defaults = data.get('defaults') or {}
        return cls(parts, edits, defaults)
    
    def to_dict(self) -> dict:
        "save StringBuilder to a dictionary of values"
        output = {'parts': self.parts}
        if self.edits:
            output['edits'] = [op.to_dict() for op in self.edits]
        return output
    
    def __call__(
        self,
        tokens: dict[str, str],
    ) -> str:
        if self.defaults:
            defaults = copy(self.defaults)
            defaults.update(tokens)
            tokens = defaults
        else:
            tokens = copy(tokens)
        tokens = {k:v for k,v in tokens.items() if v}
        for op in self.edits:
            try:
                op.modify_dict(tokens)
            except SyntaxError: # token operation failed; just skip it
                pass
        string_parts = []
        for part in self.parts:
            try:
                string_parts.append(part.format(**tokens))
            except KeyError: # skip parts that reference a nonexistent token
                pass
            # if a mandatory TokenOperation failed, don't return a URL
            except SyntaxError:
                string_parts = []
                break
        t = copy(tokens)
        return ''.join(string_parts) or None
    
    def __repr__(self):
        return (
            f'StringBuilder(parts={self.parts}'
            + (f', edits={self.edits}' if self.edits else '')
            + (f', defaults={self.defaults}' if self.defaults else '')
            + ')'
        )



# This list is needed for the 'number_style' token operation
number_words = [
    ('i', 'one', 'first'),
    ('ii', 'two', 'second'),
    ('iii', 'three', 'third'),
    ('iv', 'four', 'fourth'),
    ('v', 'five', 'fifth'),
    ('vi', 'six', 'sixth'),
    ('vii', 'seven', 'seventh'),
    ('viii', 'eight', 'eighth'),
    ('ix', 'nine', 'ninth'),
    ('x', 'ten', 'tenth'),
    ('xi', 'eleven', 'eleventh'),
    ('xii', 'twelve', 'twelfth'),
    ('xiii', 'thirteen', 'thirteenth'),
    ('xiv', 'fourteen', 'fourteenth'),
    ('xv', 'fifteen', 'fifteenth'),
    ('xvi', 'sixteen', 'sixteenth'),
    ('xvii', 'seventeen', 'seventeenth'),
    ('xviii', 'eighteen', 'eighteenth'),
    ('xix', 'nineteen', 'nineteenth'),
    ('xx', 'twenty', 'twentieth'),
    'xxi', 'xxii', 'xxiii', 'xxiv', 'xxv',
    'xxvi', 'xxvii', 'xxviii', 'xxix',
    ('xxx', 'thirty', 'thirtieth'),
    'xxxi', 'xxxii', 'xxxiii', 'xxxiv', 'xxxv',
    'xxxvi', 'xxxvii', 'xxxviii', 'xxxix',
    ('xl', 'forty', 'fortieth'),
    'xli', 'xlii', 'xliii', 'xliv', 'xlv',
    'xlvi', 'xlvii', 'xlviii', 'xlix',
    ('l', 'fifty', 'fiftieth'),
    'li', 'lii', 'liii', 'liv', 'lv',
    'lvi', 'lvii', 'lviii', 'lix',
    ('lx', 'sixty', 'sixtieth'),
    'lxi', 'lxii', 'lxiii', 'lxiv',
    'lxv', 'lxvi', 'lxvii', 'lxviii', 'lxix',
    ('lxx', 'seventy', 'seventieth'),
    'lxxi', 'lxxii', 'lxxiii', 'lxxiv', 'lxxv',
    'lxxvi', 'lxxvii', 'lxxviii', 'lxxix',
    ('lxxx', 'eighty', 'eightieth'),
    'lxxxi', 'lxxxii', 'lxxxiii', 'lxxxiv', 'lxxxv',
    'lxxxvi', 'lxxxvii', 'lxxxviii', 'lxxxix',
    ('xc', 'ninety', 'ninetieth'),
    'xci', 'xcii', 'xciii', 'xciv', 'xcv',
    'xcvi', 'xcvii', 'xcviii', 'xcix',
    ('c', 'one-hundred', 'one-hundredth'),
]
for i, entry in enumerate(number_words):
    if type(entry) is tuple:
        tens_place = entry[1]
    else:
        digit = number_words[i % 10]
        number_words[i] = (
            entry,                      # roman numeral
            f'{tens_place}-{digit[1]}', # cardinal number
            f'{tens_place}-{digit[2]}', # ordinal number
        )
number_words = tuple(number_words)



from .citation import Citation
from .citator import Citator, Template, cite, list_cites, insert_links
from .authority import Authority, list_authorities
from .tokens import TokenType, TokenOperation, StringBuilder
